extern crate serde_json;
extern crate tokio;
extern crate tokio_postgres;

mod configure;
mod parser;
mod serializer;
use tokio_postgres::types::ToSql;
use tokio_postgres::{Error, NoTls};

/* conditionals */
const FROM_FILE: bool = false;
const FROM_DEVICE: bool = true;
const NON_CHUNKED: bool = true;
const CHUNKED: bool = false;

fn query_string(insert_max: &usize, table_name: &str) -> String {
    let mut insert_template = String::with_capacity(insert_max * 8 + 96);
    insert_template.push_str(&*format!("INSERT INTO {} (packet) Values ", table_name));

    for insert in 0..insert_max - 1 {
        insert_template.push_str(&*format!("(${}), ", insert + 1));
    }
    insert_template.push_str(&*format!("(${})", insert_max));

    insert_template
}

#[tokio::main(core_threads = 4)] // Tokio is implemented for possible future use.
async fn main() -> Result<(), Error> {
    /* Init values from file */
    let config: configure::Config = configure::from_json_file().unwrap();
    let pcap_map = configure::map_pcap_dir(&config.pcap_dir).unwrap();

    // TODO: Create db table with pcap file hashes
    // TODO: hash file metadata, so its state is comparable with future file updates and can be written to a db table (and read e.g. after system crash)
    //       This db table should include UUIDs as primary keys, so it can be joined effectively with past and future runs.
    // TODO: Use inotify crate to update pcap_map according to files created while parser is running

    /* db connection */
    let (client, connection) = tokio_postgres::connect(&config.connection, NoTls).await?;
    tokio::spawn(async move {
        if let Err(e) = connection.await {
            eprintln!("connection error: {}", e);
        }
    });
    client
        .execute(&*format!("DROP TABLE IF EXISTS {}", &config.tablename), &[])
        .await?;
    client
        .execute(
            &*format!(
                "CREATE TABLE {} ( ID serial NOT NULL PRIMARY KEY, packet json NOT NULL)",
                &config.tablename
            ),
            &[],
        )
        .await?;

    /* device or file input */
    match config.is_device {
        FROM_FILE => {
            for (_pcap_file, _pcap_info) in pcap_map.iter() {
                println!("{:?}: {:?}", &_pcap_file, &_pcap_info);

                let v: Vec<parser::QryData> =
                    parser::parse(&_pcap_file, &config.filter, &config.regex_filter);
                let packets_serialized = serializer::serialize_packets(v);

                // TODO: Tuning vector capacity according to mean average & std dev of packet sizes
                //                let mut v = Vec::<parser::QryData>::with_capacity(100000);
                //                v.extend(parser::parse(&_pcap_file, &config.filter, &config.regex_filter));
                //                let mut packets_serialized = Vec::<serde_json::Value>::with_capacity(100000);
                //                packets_serialized.extend(serializer::serialize_packets(v));

                /* Do chunks and query data */
                let chunker = (&packets_serialized.len() < &config.insert_max) && (0 <  packets_serialized.len()) ;
                match chunker {
                    NON_CHUNKED => {
                        let insert_str = query_string(&packets_serialized.len(), &config.tablename);
                        let statement = client.prepare(&insert_str).await?;
                        client
                            .query_raw(
                                &statement,
                                packets_serialized.iter().map(|p| p as &dyn ToSql),
                            )
                            .await?;
                    }
                    CHUNKED => {
                        let insert_str = query_string(&config.insert_max, &config.tablename);
                        let statement = client.prepare(&insert_str).await?;

                        for chunk in packets_serialized.chunks_exact(config.insert_max) {
                            client
                                .query_raw(&statement, chunk.iter().map(|p| p as &dyn ToSql))
                                .await?;
                        }
                        let remainder_len = packets_serialized
                            .chunks_exact(config.insert_max)
                            .remainder()
                            .len();
                        if 0 < remainder_len {
                            let rem_str = query_string(&remainder_len, &config.tablename);
                            let statement = client.prepare(&rem_str).await?;
                            client
                                .query_raw(
                                    &statement,
                                    packets_serialized
                                        .chunks_exact(config.insert_max)
                                        .remainder()
                                        .iter()
                                        .map(|p| p as &dyn ToSql),
                                )
                                .await?;
                        }
                    }
                }
            }
        }
        FROM_DEVICE => {
            let insert_str = query_string(&config.insert_max, &config.tablename);
            let statement = client.prepare(&insert_str).await?;
            loop {
                let v: Vec<parser::QryData> = parser::parse_device(
                    &config.device,
                    &config.filter,
                    &config.insert_max,
                    &config.regex_filter,
                );
                let packets_serialized = serializer::serialize_packets(v);
                client
                    .query_raw(
                        &statement,
                        packets_serialized.iter().map(|p| p as &dyn ToSql),
                    )
                    .await?;
            }
        }
    }
    Ok(())
}
